"""
Normalization functions such as batch norm
"""


from __future__ import annotations
from typing import Sequence, Union, Tuple
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf


__all__ = ["moments", "LayerNorm"]


def moments(x: Tensor, axis: Union[Dim, Sequence[Dim]]) -> Tuple[Tensor, Tensor]:
    """
    :param x: input
    :param axis: the axis to be reduced, to calculate statistics over
    :return: mean, variance. it has the same shape as the input with the axis removed
    """
    mean = rf.reduce_mean(x, axis=axis)
    # stop_gradient does not change the gradient here
    variance = rf.reduce_mean(rf.squared_difference(x, rf.stop_gradient(mean)), axis=axis)
    return mean, variance


class LayerNorm(rf.Module):
    """
    `Layer normalization <https://arxiv.org/abs/1607.06450>`__.

    Note that we *just* normalize over the feature-dim axis here.
    This is consistent to the default behavior of :class:`tf.keras.layers.LayerNormalization`
    and also how it is commonly used in many models, including Transformer.

    However, there are cases where it would be common to normalize over all axes except batch-dim,
    or all axes except batch and time.
    For a more generic variant, see :func:`norm`.
    """

    def __init__(self, in_dim: Union[rf.Dim, Sequence[rf.Dim]], *, eps: float = 1e-6):
        super().__init__()
        self.in_dim = in_dim
        self.eps = eps
        self.scale = rf.Parameter([self.in_dim] if isinstance(self.in_dim, rf.Dim) else self.in_dim)
        self.scale.initial = 1.0
        self.bias = rf.Parameter(self.scale.dims)
        self.bias.initial = 0.0

    def __call__(self, x: Tensor) -> Tensor:
        mean = rf.reduce_mean(x, axis=self.in_dim)
        variance = rf.reduce_mean(rf.squared_difference(x, mean), axis=self.in_dim)
        norm_x = (x - mean) * rf.rsqrt(variance + self.eps)
        return norm_x * self.scale + self.bias



"""
Parameter / variable
"""

from __future__ import annotations
from typing import Optional, TypeVar, Sequence
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf
from ._backend import global_backend as _global_backend


__all__ = ["Parameter"]


T = TypeVar("T")


class Parameter(Tensor[T]):
    """
    This represents a (potential trainable) parameter,
    aka ``tf.Variable`` in TensorFlow,
    wrapping to ``VariableLayer`` in RETURNN.
    """

    def __init__(
        self,
        dims: Sequence[Dim],
        dtype: Optional[str] = None,
        *,
        sparse_dim: Optional[Dim] = None,
        trainable: Optional[bool] = None,
        auxiliary: bool = False,
        non_critical_for_restore: bool = False,
        weight_decay: Optional[float] = 0.0,
        initial: Optional[rf.init.ParamInitType] = None,
        raw_tensor: Optional[T] = None,
    ):
        """
        :param dims:
        :param dtype:
        :param sparse_dim:
        :param trainable: if True, and optimizer would do updates to this parameter in training mode
        :param auxiliary: if True, this indicates that this parameter should not be transformed by transformations
          such as weight normalization. One example are running statistics, as used for batch normalization.
          This usually implies that the parameter is not trainable, i.e. not to be updated by the optimizer,
          but usually has some custom update.
          This flag is not passed on to RETURNN but just used here for returnn-common logic.
        :param non_critical_for_restore: if True, this parameter is not critical for restoring a model.
        :param weight_decay:
        :param initial:
        """
        if not all(isinstance(dim, Dim) for dim in dims):
            raise TypeError(f"shape {dims} must be a sequence of Dim")
        if not all(isinstance(dim.dimension, int) for dim in dims):
            raise ValueError(f"shape {dims} must be static")
        if len(dims) != len(set((d, d.match_priority) for d in dims)):
            raise ValueError(f"shape {dims} dims must be unique")
        super(Parameter, self).__init__(
            "parameter",
            dims=dims,
            dtype=dtype or (rf.get_default_float_dtype() if not sparse_dim else rf.get_default_array_index_dtype()),
            sparse_dim=sparse_dim,
        )
        if raw_tensor is not None:
            self.raw_tensor = raw_tensor
        else:
            self.raw_tensor = _global_backend.create_parameter_raw(self)
        if auxiliary and trainable is None:
            trainable = False
        self._trainable = trainable
        self._auxiliary = auxiliary
        self._non_critical_for_restore = non_critical_for_restore
        self._weight_decay = weight_decay
        self._initial = None  # type: Optional[rf.init.ParamInitType]
        self.initial = initial

    def __copy__(self):
        # Should return new copy. https://github.com/rwth-i6/returnn_common/pull/215#issuecomment-1269651064
        # Note that the values are *not* copied, but rather it will use the same param init scheme.
        res = type(self)(
            dims=self.dims,
            dtype=self.dtype,
            trainable=self.trainable,
            auxiliary=self.auxiliary,
            non_critical_for_restore=self.non_critical_for_restore,
            weight_decay=self.weight_decay,
        )
        res.initial = self.initial
        return res

    def __deepcopy__(self, memo=None):
        # Should return new copy. https://github.com/rwth-i6/returnn_common/pull/215#issuecomment-1269651064
        # Note that the values are *not* copied, but rather it will use the same param init scheme.
        from copy import deepcopy

        res = self.__copy__()
        if isinstance(self.initial, rf.init.ParamInit):
            res.initial = deepcopy(self.initial, memo=memo)  # noqa
        else:
            res.initial = self.initial
        return res

    @property
    def initial(self) -> Optional[rf.init.ParamInitType]:
        """initial value of the parameter"""
        return self._initial

    @initial.setter
    def initial(self, value: Optional[rf.init.ParamInitType]):
        # Keep the original ParamInit, so that copies of the Parameter would have a different initial random value.
        # https://github.com/rwth-i6/returnn_common/issues/216
        self._initial = value

        if isinstance(value, rf.init.ParamInit):
            value = value(dims=self.dims, dtype=self.dtype)
        self._raw_backend.set_parameter_initial_value(self, value)

    @property
    def weight_decay(self) -> float:
        """
        Weight decay, which is equivalent to L2 loss on the parameters for SGD.
        On RETURNN side, whether this is handled separately or is part of the main loss,
        can be controlled via the ``decouple_constraints`` config option.
        https://github.com/rwth-i6/returnn_common/issues/59#issuecomment-1073913421
        """
        return self._weight_decay or 0.0

    @weight_decay.setter
    def weight_decay(self, value: Optional[float]):
        self._weight_decay = value

    @property
    def trainable(self) -> Optional[bool]:
        """trainable"""
        return self._trainable

    @trainable.setter
    def trainable(self, value: Optional[bool]):
        self._trainable = value

    @property
    def auxiliary(self) -> bool:
        """auxiliary"""
        return self._auxiliary

    @auxiliary.setter
    def auxiliary(self, value: bool):
        self._auxiliary = value

    @property
    def non_critical_for_restore(self) -> bool:
        """non_critical_for_restore"""
        return self._non_critical_for_restore

    @non_critical_for_restore.setter
    def non_critical_for_restore(self, value: bool):
        self._non_critical_for_restore = value



"""
Represents a dimension of a tensor.
A dimension can come with further information such as individual sequence lengths.

This identifies one axis/dimension, like a time-dimension, etc.
This was called ``DimensionTag`` earlier, and referred to as dimension tag.

This is used by :class:`Tensor` (earlier ``Data``). See :func:`Tensor.dims`.
This would be passed as ``dims`` when creating a :class:`Tensor` instance.

It is not to specify the specific axis in a specific Tensor,
but to specify the content and dimension.
I.e. if we have the same Dim for two Data instances,
the dimensions should match. I.e.:

    data1.dims[i] == data2.dims[j]
      =>  data1.raw_tensor.shape[i] == data2.raw_tensor.shape[j]

This also includes further information such as sequence lengths
or a vocabulary.

Deprecated: We differentiate between the batch dim, spatial dim or feature dim,
although that is just flag and in many contexts there is no real difference
between a spatial dim and a feature dim (the batch dim is often handled differently).

"""


from __future__ import annotations
from typing import Optional, Union

from ._dim_extra import _DimExtra, _DimMixin, DimTypes
from . import tensor as _t


__all__ = ["Dim", "batch_dim", "single_step_dim", "VerifyOutShapeException"]


class Dim(_DimMixin):
    """
    Represents a dimension of a tensor.
    This potentially comes with further information such as individual sequence lengths.
    See the module docstring.
    """

    Types = DimTypes  # old alias

    __slots__ = ("name", "capacity", "size", "dyn_size_ext", "_extra")

    name: Optional[str]
    capacity: Optional[int]  # shape[axis] in the raw tensor (might need power-of-two or static shape), None if dynamic
    size: Optional[int]  # shape[axis] in the represented tensor if static, None if dynamic, then dyn_size_ext
    dyn_size_ext: Optional[_t.Tensor]
    _extra: Optional[_DimExtra]

    def __init__(
        self,
        dimension: Optional[Union[int, _t.Tensor]],
        *,
        name: Optional[str] = None,
        capacity: Optional[int] = None,
        dyn_size_ext: Optional[_t.Tensor] = None,
        description: Optional[str] = None,
        **kwargs,
    ):
        # dimension is the most common way to specify whether it is static or dynamic,
        # and if dynamic, we can directly pass the dynamic sizes.
        # It also infers reasonable defaults for capacity, if this is not set explicitly.
        # This logic here also covers the old __init__ option dyn_size_ext.
        if dimension is None:
            self.capacity = capacity
            self.size = None
            self.dyn_size_ext = dyn_size_ext.copy() if dyn_size_ext else None
        elif isinstance(dimension, int):
            self.capacity = capacity or dimension
            self.size = dimension
            self.dyn_size_ext = None
        elif isinstance(dimension, _t.Tensor):
            self.capacity = capacity
            self.size = None
            self.dyn_size_ext = dimension.copy()
        else:
            raise TypeError(f"unexpected dimension type: {type(dimension)}")
        if not name and not description and self.dyn_size_ext:
            name = self.dyn_size_ext.name
        self.name = name or description
        self._extra = None

        if kwargs:
            self._handle_extra_kwargs(**kwargs)

    def __repr__(self):
        return "Dim{%s}" % self.short_repr()


# Global batch dim, which would usually be used the dataloader.
batch_dim = Dim(kind=Dim.Types.Batch, description="global batch", dimension=None)

# This indicates to perform a single step execution of some layer which can potentially have recurrent state.
single_step_dim = Dim(description="single-step", kind=Dim.Types.Spatial, special=True, dimension=1)


class VerifyOutShapeException(Exception):
    """
    Exception via :func:`Tensor.verify_out_shape`.
    """



"""
This package provides the frontend API.
The frontend API supports multiple backends.
https://github.com/rwth-i6/returnn/issues/1120

The convention for the user is to do::

    import returnn.frontend as rf
"""

# Not necessarily all here, but the most common ones.
# (Take PyTorch `torch.nn` as a reference.)

# Some most come first here when others directly use it,
# e.g. `rf.Module` as a baseclass.
from .module import *
from .state import *

# Now the rest, in alphabetical order.
from .array_ import *
from .attention import *
from .cond import *
from .const import *
from .container import *
from .dims import *
from .dropout import *
from .dtype import *
from .gradient import *
from .linear import *
from .loss import *
from .math_ import *
from .matmul import *
from .normalization import *
from .parameter import *
from .rand import *
from .reduce import *
from .run_ctx import *
from .types import *

# Modules not in the main namespace but in sub namespaces.
from . import init

# And some functions from the internal backend API.
from ._backend import select_backend_torch, select_backend_returnn_layers_tf



"""
Internal utils
"""

from __future__ import annotations
from typing import Union, Optional, Type, TypeVar, Sequence, Tuple
import numpy

from returnn import frontend as _global_rf
from returnn.frontend._backend import Backend, get_backend_by_raw_tensor_type
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf

T = TypeVar("T")


def get_backend_from_tensors(*args):
    """
    :param args:
    :return: frontend, fallback to global frontend
    """
    for x in args:
        if isinstance(x, Tensor):
            # noinspection PyProtectedMember
            return x._raw_backend
    return _global_rf


def get_dtype_name(x: Union[T, Tensor[T], int, float]) -> str:
    """
    :param x: tensor
    :return: dtype of tensor, as string
    """
    if isinstance(x, Tensor):
        return x.dtype
    elif isinstance(x, int):
        return rf.get_default_int_dtype()
    elif isinstance(x, float):
        return rf.get_default_float_dtype()
    else:
        backend = get_backend_by_raw_tensor_type(type(x))
        return backend.get_dtype_name_raw(x)


def is_int(x: Union[T, Tensor[T], int, float]) -> bool:
    """
    :param x:
    :return: whether the dtype is int
    """
    dtype = get_dtype_name(x)
    return dtype.startswith("int") or dtype.startswith("uint")


def bin_op_out_template(
    backend: Type[Backend],
    a: Union[Tensor[T], int, float, numpy.number],
    b: Union[Tensor[T], int, float, numpy.number],
    *,
    name: str,
    res_dtype: Optional[str],
    allow_broadcast_all_sources: Optional[bool] = None,
    dim_order: Optional[Sequence[Dim]] = None,
) -> Tuple[Tensor[T], Tensor[T], Tensor[T]]:
    """
    make template for output tensor of binary op

    :param backend:
    :param a:
    :param b:
    :param name: str
    :param res_dtype: if not given, infer from a and b
    :param allow_broadcast_all_sources: if True, it is allowed that neither a nor b has all dims of the result.
        Not needed when out_dims is specified explicitly.
    :param dim_order: defines the order of the resulting dims. if None, it is automatically inferred from a and b.
        Not all the dims of a and b need to be specified here, and there could also be other dims in the dim_order.
    :return: out, a, b
    """
    src_dtype = None
    if isinstance(a, Tensor):
        src_dtype = a.dtype
    elif isinstance(b, Tensor):
        src_dtype = b.dtype
    a = rf.convert_to_tensor(a, dtype=src_dtype, _backend=backend)
    src_dtype = src_dtype or a.dtype
    b = rf.convert_to_tensor(b, dtype=src_dtype, _backend=backend)
    # sanity checks
    # noinspection PyProtectedMember
    assert a._raw_backend == b._raw_backend, "Cannot combine tensors from two different frontends, e.g. TF and PT"
    if res_dtype is None:
        assert (
            a.dtype == b.dtype
        ), f"For now only operations with Tensors of the same dtypes are supported, got {a} and {b}"
    all_dims = []
    for dim in a.dims + b.dims:
        if dim in all_dims:
            continue
        # Not simply `all_dims.append(dim)`,
        # because a dim might occur multiple times in a.dims or b.dims
        # (with different match_priority),
        # e.g. in the case of square matrices.
        # Still it is the common case that they are unique,
        # and this allows for a faster path.
        if a.dims.count(dim) <= 1 and b.dims.count(dim) <= 1:
            all_dims.append(dim)
            continue
        if a.dims.count(dim) >= b.dims.count(dim):
            all_dims.extend([dim_ for dim_ in a.dims if dim_ == dim])
        else:
            all_dims.extend([dim_ for dim_ in b.dims if dim_ == dim])
    if all(set(x.dims) != set(all_dims) for x in (a, b)):
        if allow_broadcast_all_sources is False:
            raise ValueError(f"compare: sources {a!r} {b!r} not allowed with allow_broadcast_all_sources=False")
        elif allow_broadcast_all_sources is None:
            raise ValueError(f"compare: sources {a!r} {b!r} require explicit allow_broadcast_all_sources=True")
        elif allow_broadcast_all_sources is True:
            pass
        else:
            raise TypeError(f"invalid type for allow_broadcast_all_sources: {type(allow_broadcast_all_sources)}")
    if dim_order:
        all_dims.sort(key=lambda d: dim_order.index(d) if d in dim_order else len(dim_order))
    if res_dtype is None:
        res_dtype = src_dtype
    out = Tensor(name, dims=all_dims, dtype=res_dtype)
    out.feature_dim = res_feature_dim(a, b)
    a = a.copy_compatible_to(out, check_dtype=False, check_sparse=False)
    b = b.copy_compatible_to(out, check_dtype=False, check_sparse=False)
    return out, a, b


def res_feature_dim(a: Tensor, b: Tensor) -> Optional[Dim]:
    """
    :param a:
    :param b:
    :return: feature dim if consistent or None
    """
    if a.feature_dim and not b.feature_dim:
        return a.feature_dim
    if b.feature_dim and not a.feature_dim:
        return b.feature_dim
    if a.feature_dim and b.feature_dim and a.feature_dim == b.feature_dim:
        return a.feature_dim
    return None



"""
Provides needed Git utilities to manage the local checked out packages.

The packages are stored in: ``~/returnn/pkg``
(Can be configured via env ``RETURNN_PKG_PATH``)

We might have multiple checkouts of the same repo.
In this case, we want to share the object files.
We are using `git worktree <https://git-scm.com/docs/git-worktree>`__ for this.

References:

  https://github.com/golang/go/blob/724d0720/src/cmd/go/internal/modfetch/codehost/git.go
  https://golang.org/ref/mod

Note about our terminology (esp vs Go terminology):

Go differentiates between packages, modules and repositories,
and even major semantic version sub-directories inside the modules.
We do not make this distinction.
You have a repository, and some specific commit,
and the file path inside the repo.
The file path might be sth like ``transformer-v3.py``
to note that it is incompatible from an earlier version.

Note about our ``version`` (revision) format:

During development, you would want to work in a live working tree.
There is a working tree in ``...repo``, and ``version=None`` refers to that.

Once development of one piece/module/package is ready,
you always should specify the version explicitly.
The format is ``<date>-<commit>``, where ``date`` should be like ``YYYYMMDDHHMMSS``
(can potentially be shorter up to ``YYYYMMDD``)
and ``commit`` is a revision, with minimum 7 numbers of digits.

Note about our ``repo`` format:

Currently always assumed to be remote, and then cached in our repo cache.
"""

import os
import typing
from subprocess import check_call, check_output, SubprocessError
import re
from . import common

_MinNumHashDigits = 7
_DefaultNumHashDigits = 12
_FullNumHashDigits = 40
_RevDigitsRe = re.compile("^[0123456789abcdef]*$")
_DateFormat = "%Y%m%d%H%M%S"
_DateFormatLen = 14
_DateFormatAllowedLens = (8, 10, 12, 14)  # Day only, or partial time.
# Not too complicated but also not totally dumb.
_DateRe = re.compile(r"^(\d\d\d\d)([01]\d)([0123]\d)(([012]\d)(([0-5]\d)([0-5]\d)?)?)?$")


def stat_repo(repo, version):
    """
    :param str repo: e.g. "github.com/rwth-i6/returnn-experiments"
    :param str|None version: e.g. "20211231-0123abcd0123"
    """
    repo_ = _get_repo(repo)
    repo_.get_work_dir(version)


def get_repo_path(repo, version, _report_dev_version_usage_stack_frame_depth=1):
    """
    :param str repo: e.g. "github.com/rwth-i6/returnn-experiments"
    :param str|None version: e.g. "20211231-0123abcd0123"
    :param int _report_dev_version_usage_stack_frame_depth:
    :return: path to repo
    :rtype: str
    """
    repo_ = _get_repo(repo)
    work_dir = repo_.get_work_dir(version)
    repo_path = work_dir.get_path()
    if not version:
        _report_usage_dev_version(
            repo_path=repo_path, stack_frame_depth=_report_dev_version_usage_stack_frame_depth + 1
        )
    return repo_path


def _simple_validate_repo_name(repo):
    """
    :param str repo:
    """
    assert ".." not in repo and ":" not in repo


def _main_repo_path(repo):
    """
    :param str repo:
    :return: main repo dir (which includes the Git objects)
    :rtype: str
    """
    return "%s/%s" % (common.package_path(), repo)


def _dev_repo_path(repo):
    """
    :param str repo:
    :return: dev working tree of a repo. currently the same as the main repo path
    :rtype: str
    """
    return _main_repo_path(repo)


def _repo_path(repo, version):
    """
    :param str repo:
    :param str|None version:
    :rtype: str
    """
    if not version:
        return _dev_repo_path(repo)
    return "%s@v%s" % (_main_repo_path(repo), version)


def _repo_remote_url(repo):
    """
    :param str repo:
    :rtype: str
    """
    # Note: See also Go documentation how it does this: https://golang.org/ref/mod#vcs-find
    # It uses https by default.
    # You can overwrite this via the global Git configuration.
    # Example for GitHub: put this into your ~/.gitconfig:
    # [url "git@github.com:"]
    #     insteadOf = https://github.com/
    p = repo.find("/")
    assert p >= 0
    host, path = repo[:p], repo[p:]
    return "https://%s:%s" % (host, path)


def _simple_validate_commit_rev(rev):
    """
    :param str rev:
    """
    assert len(rev) >= _MinNumHashDigits and _RevDigitsRe.match(rev)


def _simple_validate_date(date):
    """
    :param str date:
    """
    assert len(date) in _DateFormatAllowedLens and _DateRe.match(date)


def _simple_validate_version(version):
    """
    :param str|None version:
    """
    if not version:
        return
    date, rev = version.split("-")
    _simple_validate_date(date)
    _simple_validate_commit_rev(rev)


def _rev_from_version(version):
    """
    :param str version: e.g. "20211231-0123abcd0123"
    :return: e.g. "0123abcd0123"
    :rtype: str
    """
    p = version.rfind("-")
    if p < 0:  # ok, allow here
        _simple_validate_commit_rev(version)
        return version
    rev = version[p + 1 :]
    _simple_validate_commit_rev(rev)
    return rev


def _version_from_date_and_rev(date, rev):
    """
    :param str date:
    :param str rev:
    """
    _simple_validate_commit_rev(rev)
    return "%s-%s" % (date, rev)


def _sys_git_clone_repo(repo):
    """
    :param str repo:
    """
    url = _repo_remote_url(repo)
    main_path = _main_repo_path(repo)
    common.logger.info("Git clone %s", repo)
    check_call(["git", "clone", url, main_path])


def _sys_git_fetch(repo):
    """
    :param str repo:
    """
    # Note: Just simply git fetch, no extra logic.
    # We might want to allow for named refs later...
    # We might want to extend this logic to be more like Go...
    # https://github.com/golang/go/blob/724d0720b3e/src/cmd/go/internal/modfetch/codehost/git.go#L366
    main_path = _main_repo_path(repo)
    common.logger.info("Git fetch %s", repo)
    check_call(["git", "fetch"], cwd=main_path)


def _sys_git_create_repo_workdir(repo, version):
    """
    :param str repo:
    :param str version:
    """
    main_path = _main_repo_path(repo)
    rev = _rev_from_version(version)
    version_path = _repo_path(repo, version)
    common.logger.info("Git add worktree %s: %s", repo, version)
    check_call(["git", "worktree", "add", version_path, rev], cwd=main_path)


def _sys_git_stat_local_rev(repo, rev):
    """
    :param str repo:
    :param str rev:
    :return: (full_rev, date)
    :rtype: (str|None, str|None)
    """
    main_path = _main_repo_path(repo)
    try:
        out = check_output(
            [
                "git",
                "-c",
                "log.showsignature=false",
                "log",
                "-n1",
                "--format=format:%H %cd",
                "--date=format:%s" % _DateFormat,
                rev,
                "--",
            ],
            cwd=main_path,
        )
    except SubprocessError:
        return None, None
    out = out.decode("utf8")
    full_rev, date = out.split()
    assert full_rev.startswith(rev) and len(full_rev) == _FullNumHashDigits
    _simple_validate_commit_rev(full_rev)
    _simple_validate_date(date)
    return full_rev, date


class _Repo:
    def __init__(self, name):
        """
        :param str name: e.g. "github.com/rwth-i6/returnn-experiments"
        """
        _simple_validate_repo_name(name)
        self.name = name
        self._cloned = False
        self._loaded_local_work_dirs = False
        self._dev_work_dir = None  # type: typing.Optional[_RepoWorkDir]
        self._work_dirs = {}  # type: typing.Dict[str,_RepoWorkDir]  # commit-rev -> entry

    def get_dev_dir_path(self):
        """
        :rtype: str
        """
        self._clone()
        return _dev_repo_path(self.name)

    def _clone(self):
        if self._cloned:
            return
        main_path = _main_repo_path(self.name)
        if not os.path.exists(main_path):
            _sys_git_clone_repo(self.name)
            assert os.path.exists(main_path)
        self._cloned = True

    def _load_work_dirs(self):
        if self._loaded_local_work_dirs:
            return
        self._clone()
        main_path = _main_repo_path(self.name)
        p = main_path.rfind("/")
        assert p > 0
        dir_name, base_name = main_path[:p], main_path[p + 1 :]
        prefix = base_name + "@v"
        for d in sorted(os.listdir(dir_name)):
            if not d.startswith(prefix):
                continue
            version = d[len(prefix) :]
            rev = _rev_from_version(version)
            if rev in self._work_dirs:
                continue
            self._work_dirs[rev] = _RepoWorkDir(repo=self, version=version)
        self._loaded_local_work_dirs = True

    def _get_work_dir_stat_local(self, rev):
        """
        :param str rev:
        :rtype: _RepoWorkDir
        """
        full_rev, date = _sys_git_stat_local_rev(self.name, rev)
        if full_rev:
            default_rev = full_rev[:_DefaultNumHashDigits]
            assert default_rev not in self._work_dirs
            default_version = _version_from_date_and_rev(date=date, rev=default_rev)
            _sys_git_create_repo_workdir(self.name, version=default_version)
            work_dir = _RepoWorkDir(self, default_version)
            self._work_dirs[default_rev] = work_dir
            return work_dir
        return None

    def get_work_dir(self, version):
        """
        :param str|None version:
        :rtype: _RepoWorkDir
        :return: work dir. this makes sure that the work dir also exists
        """
        if not version:
            if self._dev_work_dir:
                return self._dev_work_dir
            self._clone()
            self._dev_work_dir = _RepoWorkDir(self, None)
            return self._dev_work_dir
        self._load_work_dirs()
        rev = _rev_from_version(version)
        if rev in self._work_dirs:
            work_dir = self._work_dirs[rev]
            work_dir.validate_version(version)
            return work_dir
        for rev_, work_dir in self._work_dirs.items():
            if rev_.startswith(rev) or (len(rev) > len(rev_) and rev.startswith(rev_)):
                work_dir.validate_version(version)
                self._work_dirs[rev] = work_dir
                return work_dir
        work_dir = self._get_work_dir_stat_local(rev)
        if work_dir:
            work_dir.validate_version(version)
            return work_dir
        _sys_git_fetch(self.name)
        work_dir = self._get_work_dir_stat_local(rev)
        assert work_dir, "Git repo %s, version %s unknown." % (self.name, version)
        work_dir.validate_version(version)
        return work_dir


class _RepoWorkDir:
    def __init__(self, repo, version):
        """
        :param str|_Repo repo:
        :param str|None version: (normalized)
        """
        if not isinstance(repo, _Repo):
            repo = _get_repo(repo)
        _simple_validate_version(version)
        self.repo = repo
        self.version = version
        if version:
            self.version_date, self.version_rev = version.split("-")
        else:
            self.version_date, self.version_rev = None, None

    def get_path(self):
        """
        :rtype: str
        """
        return _repo_path(self.repo.name, self.version)

    def validate_version(self, version):
        """
        :param str|None version:
        """
        if version:
            assert self.version
            date, rev = version.split("-")
            _simple_validate_date(date)
            _simple_validate_commit_rev(rev)
            if not self.version_date.startswith(date):
                raise common.InvalidVersion("Version %s, date does not match in %s" % (self.version, version))
            if len(rev) <= len(self.version_rev):
                if not self.version_rev.startswith(rev):
                    raise common.InvalidVersion("Version %s, revision does not match in %s" % (self.version, version))
            else:
                if not rev.startswith(self.version_rev):
                    raise common.InvalidVersion("Version %s, revision does not match in %s" % (self.version, version))
        else:
            if self.version:
                raise common.InvalidVersion("Development version, got %s" % (version,))


_repo_cache = {}  # type: typing.Dict[str,_Repo]


def _get_repo(repo):
    """
    :param str repo:
    :rtype: _Repo
    """
    assert isinstance(repo, str)
    obj = _repo_cache.get(repo)
    if obj:
        return obj
    obj = _Repo(repo)
    _repo_cache[repo] = obj
    return obj


def _report_usage_dev_version(repo_path, stack_frame_depth):
    """
    :param str repo_path:
    :param int stack_frame_depth:
    """
    common.logger.warn("Access to development working tree: %s", repo_path)
    from returnn.util.basic import try_get_stack_frame

    frame = try_get_stack_frame(depth=stack_frame_depth + 1)
    if frame:
        from returnn.util.better_exchook import get_source_code, add_indent_lines

        src = get_source_code(filename=frame.f_code.co_filename, lineno=frame.f_lineno, module_globals=frame.f_globals)
        common.logger.warn(
            "  Called from: %s:%i, code:\n%s",
            frame.f_code.co_filename,
            frame.f_lineno,
            add_indent_lines("    ", src.rstrip()),
        )
    else:
        common.logger.warn("  (Could not get stack frame information from calling code.)")
    from returnn.util.basic import git_commit_date, git_commit_rev, git_is_dirty

    rev = git_commit_rev(git_dir=repo_path, length=_DefaultNumHashDigits)
    commit_date = git_commit_date(git_dir=repo_path)  # like "20190202.154527"
    commit_date = commit_date.replace(".", "")
    version = "%s-%s" % (commit_date[:8], rev[:_MinNumHashDigits])
    import_str = "v%s_%s" % (commit_date, rev)
    common.logger.warn("  Current version: %s (%s)", version, import_str)
    if git_is_dirty(git_dir=repo_path):
        common.logger.warn("  (Warning, code is dirty. Commit your recent changes.)")



"""
Like SprintErrorSignals.py but for TensorFlow.
"""

from returnn.sprint.error_signals import SprintInstancePool
import tensorflow as tf
import returnn.tf.compat as tf_compat


def py_get_sprint_automata_for_batch(sprint_opts, tags):
    """
    :param dict[str] sprint_opts:
    :param list[str] tags:
    :return: (edges, weights, start_end_states)
    :rtype: (numpy.ndarray, numpy.ndarray, numpy.ndarray)
    """
    # Also see :class:`SprintAlignmentAutomataOp`.
    sprint_instance_pool = SprintInstancePool.get_global_instance(sprint_opts=sprint_opts)
    with sprint_instance_pool.lock:  # We need multi-threading safety.
        edges, weights, start_end_states = sprint_instance_pool.get_automata_for_batch(tags)
    # Note: UnimplementedError: Unsupported numpy type 6 (uint32) -> cast to int32.
    edges = edges.astype("int32")
    start_end_states = start_end_states.astype("int32")
    return edges, weights, start_end_states


def get_sprint_automata_for_batch_op(sprint_opts, tags):
    """
    :param dict[str] sprint_opts:
    :param tf.Tensor tags: shape (batch,), of dtype string
    :return: (edges, weights, start_end_states). all together in one automaton.
      edges are of shape (4, num_edges), each (from, to, emission-idx, seq-idx), of dtype int32.
      weights are of shape (num_edges,), of dtype float32.
      start_end_states are of shape (2, batch), each (start,stop) state idx, batch = len(tags), of dtype int32.
    :rtype: (tf.Tensor, tf.Tensor, tf.Tensor)
    """

    def py_wrap_get_sprint_automata_for_batch(py_tags):
        """
        :param list[str] py_tags: len batch
        :return: (edges, weights, start_end_states)
        :rtype: (numpy.ndarray, numpy.ndarray, numpy.ndarray)
        """
        try:
            return py_get_sprint_automata_for_batch(sprint_opts=sprint_opts, tags=py_tags)
        except Exception:
            print("Exception in py_wrap_get_sprint_automata_for_batch:")
            import sys

            sys.excepthook(*sys.exc_info())
            raise

    tags.set_shape((None,))  # (batch,)
    edges, weights, start_end_states = tf_compat.v1.py_func(
        py_wrap_get_sprint_automata_for_batch,
        [tags],
        [tf.int32, tf.float32, tf.int32],
        name="get_sprint_automata_for_batch",
    )
    assert isinstance(edges, tf.Tensor)
    assert isinstance(weights, tf.Tensor)
    assert isinstance(start_end_states, tf.Tensor)
    edges.set_shape((4, None))  # (4, num_edges)
    weights.set_shape((None,))  # (num_edges,)
    start_end_states.set_shape((2, tags.get_shape().dims[0]))  # (2, batch)
    return edges, weights, start_end_states


def py_get_sprint_loss_and_error_signal(sprint_opts, log_posteriors, seq_lengths, seq_tags):
    """
    :param dict[str] sprint_opts:
    :param numpy.ndarray log_posteriors: 3d (time,batch,label)
    :param numpy.ndarray seq_lengths: 1d (batch)
    :param list[str] seq_tags: seq names
    :return: (loss, error_signal), error_signal has the same shape as posteriors. loss is a 1d-array (batch).
    :rtype: (numpy.ndarray, numpy.ndarray)
    """
    # Also see :class:`SprintErrorSigOp`.
    sprint_instance_pool = SprintInstancePool.get_global_instance(sprint_opts=sprint_opts)
    with sprint_instance_pool.lock:  # We need multi-threading safety.
        loss, error_signal = sprint_instance_pool.get_batch_loss_and_error_signal(
            log_posteriors=log_posteriors, seq_lengths=seq_lengths, tags=seq_tags
        )
    return loss, error_signal


def get_sprint_loss_and_error_signal(sprint_opts, log_posteriors, seq_lengths, seq_tags):
    """
    :param dict[str] sprint_opts:
    :param tf.Tensor log_posteriors: 3d (time,batch,label)
    :param tf.Tensor seq_lengths: 1d (batch,)
    :param tf.Tensor seq_tags: 1d (batch,), seq names
    :return: (loss, error_signal), error_signal has the same shape as posteriors. loss is a 1d-array (batch).
    :rtype: (tf.Tensor, tf.Tensor)
    """

    def py_wrap_get_sprint_loss_and_error_signal(py_log_posteriors, py_seq_lengths, py_seq_tags):
        """
        :param numpy.ndarray py_log_posteriors: 3d (time,batch,label)
        :param numpy.ndarray py_seq_lengths: 1d (batch)
        :param list[str] py_seq_tags:
        :return: (loss, error_signal), error_signal has the same shape as posteriors. loss is a 1d-array (batch).
        :rtype: (numpy.ndarray, numpy.ndarray)
        """
        try:
            return py_get_sprint_loss_and_error_signal(
                sprint_opts=sprint_opts,
                log_posteriors=py_log_posteriors,
                seq_lengths=py_seq_lengths,
                seq_tags=py_seq_tags,
            )
        except Exception:
            print("Exception in py_wrap_get_sprint_loss_and_error_signal:")
            import sys

            sys.excepthook(*sys.exc_info())
            raise

    log_posteriors.set_shape((None, None, None))  # (time,batch,label)
    seq_lengths.set_shape((None,))  # (batch,)
    seq_tags.set_shape((None,))  # (batch,)
    loss, error_signal = tf_compat.v1.py_func(
        py_wrap_get_sprint_loss_and_error_signal,
        [log_posteriors, seq_lengths, seq_tags],
        [tf.float32, tf.float32],
        name="get_sprint_loss_and_error_signal",
    )
    assert isinstance(loss, tf.Tensor)
    assert isinstance(error_signal, tf.Tensor)
    loss.set_shape((None,))  # (batch,)
    error_signal.set_shape(log_posteriors.get_shape())  # (time,batch,label)
    return loss, error_signal



"""
Dropout
"""

from __future__ import annotations
from typing import Union, Sequence
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf


__all__ = ["dropout"]


def dropout(
    source: Tensor,
    drop_prob: Union[float, Tensor],
    *,
    axis: Union[Dim, Sequence[Dim]],
    on_forward: bool = False,
) -> Tensor:
    """
    Applies dropout.

    Dropout will only be applied during training (unless you set on_forward=True).

    When dropout is applied, the output will be scaled by 1/dropout.

    :param source:
    :param drop_prob: 0.0 means to apply no dropout. 100% would mask everything.
        For every value in the tensor, the probability of it being dropped
        is drawn independently given this probability.
        The broadcasted axes are those not specified in ``axis``.
    :param axis: axis to apply dropout on. multiple axes can be specified.
        This defines the set of axes where the dropout mask is not broadcasted to.
        (RETURNN also has the ``noise_shape`` option but the ``axis`` option provides the same functionality.)
    :param on_forward: apply dropout during inference and training (so just always). otherwise only during training.
    """
    keep_prob = 1.0 - drop_prob
    if isinstance(axis, Dim):
        noise_dims = (axis,)
    else:
        noise_dims = axis
    if not set(noise_dims).issubset(source.dims):
        raise ValueError(f"dropout axis {axis} not in source {source}")

    if isinstance(keep_prob, (float, int)) and not 0 < keep_prob <= 1:
        raise ValueError("keep_prob must be a scalar tensor or a float in the " "range (0, 1], got %g" % keep_prob)

    # Do nothing if we know keep_prob == 1
    if isinstance(keep_prob, (float, int)) and keep_prob == 1:
        return source

    if on_forward:
        return _dropout(source, keep_prob, noise_dims=noise_dims)

    return rf.cond(
        pred=rf.get_run_ctx().train_flag,
        true_fn=lambda: _dropout(source, keep_prob, noise_dims=noise_dims),
        false_fn=lambda: source,
    )


def _dropout(
    x: Tensor, keep_prob: Union[float, Tensor], noise_dims: Sequence[Dim], seed=None, apply_correction_factor=True
) -> Tensor:
    """
    Computes dropout.

    Adopted from tf_util.dropout.
    Like :func:`tf.nn.dropout` but avoid :func:`tf.div` if possible.

    Note that in tf_util.dropout, we had special logic for recurrent loops:
    The mask would be created outside the loop
    and then the same mask would be used for every frame of the loop.
    We cannot really support such a logic for eager-based frameworks.

    :param x:
    :param keep_prob:
    :param noise_dims: other dims would broadcast
    :param seed: passed on to :func:`random` for the mask
    :param bool apply_correction_factor:
    """
    # uniform [keep_prob, 1.0 + keep_prob)
    random_tensor = keep_prob + rf.random_uniform(dims=noise_dims, seed=seed, dtype=x.dtype, minval=0.0, maxval=1.0)
    # 0. if [keep_prob, 1.0) and 1. if [1.0, 1.0 + keep_prob)
    binary_tensor = rf.floor(random_tensor)
    if apply_correction_factor:
        # Apply the factor on binary_tensor, because that is potentially smaller than x.
        # Use `*(1/p)` instead of `/p` because that might be faster in some cases.
        binary_tensor *= 1.0 / keep_prob

    ret = x * binary_tensor
    return ret



"""
Attention
"""


from __future__ import annotations
from typing import Tuple, Union, Optional, Sequence
from returnn.util.py_compat import Protocol
from returnn.tensor import Tensor, Dim, single_step_dim
import returnn.frontend as rf


__all__ = [
    "AttentionFunc",
    "dot_attention",
    "SelfAttentionBase",
    "SelfAttention",
    "CausalSelfAttention",
    "CausalSelfAttentionState",
]


class AttentionFunc(Protocol):
    """Protocol defining a generic attention function"""

    def __call__(
        self,
        query: Tensor,
        keys: Tensor,
        values: Tensor,
        *,
        key_dim: Dim,
        axis: Dim,
        att_dropout: float = 0.1,
    ):
        ...


def dot_attention(
    query: Tensor, keys: Tensor, values: Tensor, *, key_dim: Dim, axis: Dim, att_dropout: float = 0.0
) -> Tensor:
    """
    Calculates attention over the given axis, for given key dim.
    Any other unrelated axes do not matter here.
    This can be used for multi-head or single head.
    The query can have other dimensions or not.

    :param query: {..., key_dim}. For self-attention, do not use the `axis` as in `keys` and `values`,
        but rather replace it by another new dim via :func:`replace_dim`.
    :param keys: {..., axis, key_dim}
    :param values: {..., axis}
    :param key_dim: dim in keys and query, to be reduced to calculate the attention energies.
    :param axis: in keys and values, to apply attention on. softmax will be over this axis, and then it will be reduced
    :param att_dropout: dropout for attention weights
    :return: like values but with axis removed, and maybe any additional axes from query
    """
    query *= key_dim.dimension**-0.5
    energy = rf.matmul(query, keys, reduce=key_dim)
    att_weights = rf.softmax(energy, axis=axis)
    att_weights = rf.dropout(att_weights, att_dropout, axis=axis)
    # Masking not needed because softmax should already have masked,
    # so we have 0.0 att weights for padded frames.
    att = rf.matmul(att_weights, values, reduce=axis, disable_masking=True)
    if values.feature_dim in att.dims:
        att.feature_dim = values.feature_dim
    return att


# noinspection PyAbstractClass
class SelfAttentionBase(rf.Module):
    """
    Shared base class for (non-causal) self attention (:class:`SelfAttention`)
    and causal self attention (:class:`CausalSelfAttention`).

    It uses :func:`dot_attention` for multi-headed dot-attention.
    """

    def __init__(
        self,
        in_dim: Dim,
        proj_dim: Optional[Dim],
        *,
        key_dim_total: Dim,
        value_dim_total: Dim,
        num_heads: Union[int, Dim],
        with_bias: bool = True,
        att_dropout: float = 0.1,
    ):
        """
        :param in_dim: input dim
        :param proj_dim: if given, will add a final linear projection to this dim.
          otherwise no projection after the attention
        :param key_dim_total: total key dim. should be a multiple of num_heads
        :param value_dim_total: total value dim. should be a multiple of num_heads
        :param num_heads: number of heads
        :param with_bias: whether to add bias to qkv and proj linear projections.
          Was False in original Transformer, but many recent implementations use True by default.
          Also see: https://github.com/rwth-i6/returnn_common/issues/234.
        :param att_dropout: dropout for attention weights
        """
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = proj_dim if proj_dim else value_dim_total
        if isinstance(num_heads, int):
            num_heads = Dim(num_heads, name="num_heads")
        self.key_dim_total = key_dim_total
        self.key_dim_per_head = key_dim_total.div_left(num_heads)
        self.value_dim_total = value_dim_total
        self.value_dim_per_head = value_dim_total.div_left(num_heads)
        self.num_heads = num_heads
        self.qkv_dim_total = 2 * key_dim_total + value_dim_total
        self.qkv_dim_per_head = 2 * self.key_dim_per_head + self.value_dim_per_head
        self.qkv = rf.Linear(in_dim, self.qkv_dim_total, with_bias=with_bias)
        if proj_dim:
            self.proj = rf.Linear(value_dim_total, proj_dim, with_bias=with_bias)
        else:
            self.proj = None
        self.att_dropout = att_dropout

    def forward_qkv(self, source: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        """
        :return: q,k,v
        """
        qkv = self.qkv(source)
        qkv = rf.split_dims(qkv, axis=self.qkv_dim_total, dims=(self.num_heads, self.qkv_dim_per_head))
        q, k, v = rf.split(
            qkv,
            axis=self.qkv_dim_per_head,
            out_dims=(self.key_dim_per_head, self.key_dim_per_head, self.value_dim_per_head),
        )
        return q, k, v

    def attention(self, q: Tensor, k: Tensor, v: Tensor, *, kv_axis: Dim) -> Tensor:
        """apply attention"""
        att = dot_attention(q, k, v, key_dim=self.key_dim_per_head, axis=kv_axis, att_dropout=self.att_dropout)
        output, _ = rf.merge_dims(att, dims=(self.num_heads, self.value_dim_per_head), out_dim=self.value_dim_total)
        if self.proj:
            output = self.proj(output)
        return output


class SelfAttention(SelfAttentionBase):
    """
    Classic self attention on sequence level
    """

    def __call__(self, source: Tensor, *, axis: Dim) -> Tensor:
        """forward"""
        q, k, v = self.forward_qkv(source)
        kv_axis = Dim(None, name=f"{axis.name}-kv")
        k, _ = rf.replace_dim(k, in_dim=axis, out_dim=kv_axis)
        v, _ = rf.replace_dim(v, in_dim=axis, out_dim=kv_axis)
        return self.attention(q, k, v, kv_axis=kv_axis)


class CausalSelfAttention(SelfAttentionBase):
    """
    Classic causal self attention
    """

    def __call__(
        self,
        source: Tensor,
        axis: Dim,
        *,
        state: CausalSelfAttentionState,
    ) -> Tuple[Tensor, CausalSelfAttentionState]:
        """forward"""
        assert axis == single_step_dim  # not implemented otherwise currently...
        q, k, v = self.forward_qkv(source)
        assert state
        hist_dim = Dim(None, name="kv-history")
        new_state = CausalSelfAttentionState()
        k, _ = rf.cum_concat_step(k, prev_accum=state.k_accum, out_spatial_dim=hist_dim, axis=state.accum_axis)
        v, _ = rf.cum_concat_step(v, prev_accum=state.v_accum, out_spatial_dim=hist_dim, axis=state.accum_axis)
        new_state.k_accum = k
        new_state.v_accum = v
        new_state.accum_axis = hist_dim
        output = self.attention(q, k, v, kv_axis=hist_dim)
        return output, new_state

    def default_initial_state(self, *, batch_dims: Sequence[Dim]) -> CausalSelfAttentionState:
        """
        For causal attention.
        """
        # Note: This dim tag is wrong. It should match to the expand_dim inside __call__.
        # So the dim tag itself should be part of the layer state, and we need to define the initial value of it here.
        # This is not really supported, in various ways, also including RETURNN.
        # We just keep this code in place to be prepared for that.
        # The reason it works right now is that we do an optimization where we replace zero init state by 0.
        expand_dim = Dim(0, name="self_att_expand_dim_init")
        return CausalSelfAttentionState(
            k_accum=rf.zeros(list(batch_dims) + [expand_dim, self.num_heads, self.key_dim_per_head]),
            v_accum=rf.zeros(list(batch_dims) + [expand_dim, self.num_heads, self.value_dim_per_head]),
            accum_axis=expand_dim,
        )


class CausalSelfAttentionState(rf.State):
    """
    State for :class:`StepwiseCausalSelfAttention`.
    """

    def __init__(self, *, k_accum: Tensor = None, v_accum: Tensor = None, accum_axis: Dim = None):
        """
        :param k_accum: accumulated keys
        :param v_accum: accumulated values
        :param accum_axis:
        """
        super().__init__()
        self.k_accum = k_accum
        self.v_accum = v_accum
        self.accum_axis = accum_axis



"""
Conditional execution
"""

from __future__ import annotations
from typing import Union
from returnn.tensor import Tensor


def cond(pred: Union[bool, Tensor], true_fn, false_fn):
    """
    :param pred:
    :param true_fn:
    :param false_fn:
    :return: true_fn() if pred else false_fn()
    """
    if isinstance(pred, bool):
        if pred:
            return true_fn()
        else:
            return false_fn()
    # noinspection PyProtectedMember
    backend = pred._raw_backend
    if backend.executing_eagerly():
        if bool(pred.raw_tensor):
            return true_fn()
        else:
            return false_fn()
    raise NotImplementedError("cond() not implemented for backend %r" % backend)



"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
import _setup_test_env  # noqa
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim
from rf_utils import run_model


def test_dot_attention():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    key_dim = Dim(7, name="key")
    value_dim = Dim(13, name="value")
    extern_data = TensorDict(
        {
            "q": Tensor("q", [batch_dim, time_dim, key_dim], dtype="float32"),
            "k": Tensor("k", [batch_dim, time_dim, key_dim], dtype="float32"),
            "v": Tensor("v", [batch_dim, time_dim, value_dim], dtype="float32", feature_dim_axis=2),
        }
    )

    class _Net(rf.Module):
        def __call__(self, q: Tensor, k: Tensor, v: Tensor) -> Tensor:
            kv_axis = Dim(None, name=f"kv-axis")
            k, _ = rf.replace_dim(k, in_dim=time_dim, out_dim=kv_axis)
            v, _ = rf.replace_dim(v, in_dim=time_dim, out_dim=kv_axis)
            return rf.dot_attention(q, k, v, axis=kv_axis, key_dim=key_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(q=extern_data["q"], k=extern_data["k"], v=extern_data["v"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, value_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)



"""
Dot / matmul
"""


from __future__ import annotations
from typing import Sequence, Union, TypeVar
from returnn.tensor import Tensor, Dim

T = TypeVar("T")

__all__ = ["matmul", "dot"]


# noinspection PyShadowingNames
def matmul(
    a: Tensor[T], b: Tensor[T], *, reduce: Union[Dim, Sequence[Dim]], disable_masking: bool = False
) -> Tensor[T]:
    """
    This performs a batched matmul of two sources a and b
    (non-batched matmul and dot product are special cases).
    The underlying operation is a batched matmul (shared..., I, J) * (shared..., J, K) -> (shared..., I, K).
    The inputs a and b are transformed internally into the required shapes in the following way:
    The axis J is specified via the Dim given as 'reduce'. If multiple reduce Dims are given the corresponding axes
    are merged into one before the matmul via a reshape. All other matching Dims in a and b will be treated as
    batch dimensions ('shared...'). Dims unique to a and b define the axes I and K, respectively. (Multiple or no
    unique axes in a and b are supported too.)

    Depending on which Dims exist in a, b and reduce this dot operation can be used to compute scaling, scalar
    product, outer product, matrix-vector multiplication, matrix-matrix multiplication etc. (all possibly batched).

    :param a:
    :param b:
    :param reduce: Dims over which to perform the product, have to be present in both a and b
    :param disable_masking: If the reduction is over dynamic axes, to get the correct sum reduction,
        we need to apply masking to one of the inputs. This is done automatically.
        By enabling this flag, this would be disabled.
    :return: result of dot product, Dim order: common axes as sorted in a, unique axes of a (in order),
        unique axes of b (in order)
    """
    # noinspection PyProtectedMember
    return a._raw_backend.matmul(a=a, b=b, reduce=reduce, disable_masking=disable_masking)


# alias for some older code
dot = matmul
